@@ -264,33 +264,20 @@ xkb_filter_new(struct xkb_state *state)
 
 /***====================================================================***/
 
-static bool
-xkb_filter_group_set_func(struct xkb_state *state,
-                          struct xkb_filter *filter,
-                          const struct xkb_key *key,
-                          enum xkb_key_direction direction)
-{
-    if (key != filter->key) {
-        filter->action.group.flags &= ~ACTION_LOCK_CLEAR;
-        return true;
-    }
-
-    if (direction == XKB_KEY_DOWN) {
-        filter->refcnt++;
-        return false;
-    }
-    else if (--filter->refcnt > 0) {
-        return false;
-    }
-
-    state->components.base_group = filter->priv;
-
-    if (filter->action.group.flags & ACTION_LOCK_CLEAR)
-        state->components.locked_group = 0;
-
-    filter->func = NULL;
-    return true;
-}
+enum xkb_filter_result {
+    /*
+     * The event is consumed by the filters.
+     *
+     * An event is always processed by all filters, but any filter can
+     * prevent it from being processed further by consuming it.
+     */
+    XKB_FILTER_CONSUME,
+    /*
+     * The event may continue to be processed as far as this filter is
+     * concerned.
+     */
+    XKB_FILTER_CONTINUE,
+};
 
 static void
 xkb_filter_group_set_new(struct xkb_state *state, struct xkb_filter *filter)
@@ -303,23 +290,31 @@ xkb_filter_group_set_new(struct xkb_state *state, struct xkb_filter *filter)
 }
 
 static bool
-xkb_filter_group_lock_func(struct xkb_state *state,
-                           struct xkb_filter *filter,
-                           const struct xkb_key *key,
-                           enum xkb_key_direction direction)
+xkb_filter_group_set_func(struct xkb_state *state,
+                          struct xkb_filter *filter,
+                          const struct xkb_key *key,
+                          enum xkb_key_direction direction)
 {
-    if (key != filter->key)
-        return true;
+    if (key != filter->key) {
+        filter->action.group.flags &= ~ACTION_LOCK_CLEAR;
+        return XKB_FILTER_CONTINUE;
+    }
 
     if (direction == XKB_KEY_DOWN) {
         filter->refcnt++;
-        return false;
+        return XKB_FILTER_CONSUME;
     }
-    if (--filter->refcnt > 0)
-        return false;
+    else if (--filter->refcnt > 0) {
+        return XKB_FILTER_CONSUME;
+    }
+
+    state->components.base_group = filter->priv;
+
+    if (filter->action.group.flags & ACTION_LOCK_CLEAR)
+        state->components.locked_group = 0;
 
     filter->func = NULL;
-    return true;
+    return XKB_FILTER_CONTINUE;
 }
 
 static void
@@ -332,30 +327,23 @@ xkb_filter_group_lock_new(struct xkb_state *state, struct xkb_filter *filter)
 }
 
 static bool
-xkb_filter_mod_set_func(struct xkb_state *state,
-                        struct xkb_filter *filter,
-                        const struct xkb_key *key,
-                        enum xkb_key_direction direction)
+xkb_filter_group_lock_func(struct xkb_state *state,
+                           struct xkb_filter *filter,
+                           const struct xkb_key *key,
+                           enum xkb_key_direction direction)
 {
-    if (key != filter->key) {
-        filter->action.mods.flags &= ~ACTION_LOCK_CLEAR;
-        return true;
-    }
+    if (key != filter->key)
+        return XKB_FILTER_CONTINUE;
 
     if (direction == XKB_KEY_DOWN) {
         filter->refcnt++;
-        return false;
+        return XKB_FILTER_CONSUME;
     }
-    else if (--filter->refcnt > 0) {
-        return false;
-    }
-
-    state->clear_mods = filter->action.mods.mods.mask;
-    if (filter->action.mods.flags & ACTION_LOCK_CLEAR)
-        state->components.locked_mods &= ~filter->action.mods.mods.mask;
+    if (--filter->refcnt > 0)
+        return XKB_FILTER_CONSUME;
 
     filter->func = NULL;
-    return true;
+    return XKB_FILTER_CONTINUE;
 }
 
 static void
@@ -365,27 +353,30 @@ xkb_filter_mod_set_new(struct xkb_state *state, struct xkb_filter *filter)
 }
 
 static bool
-xkb_filter_mod_lock_func(struct xkb_state *state,
-                         struct xkb_filter *filter,
-                         const struct xkb_key *key,
-                         enum xkb_key_direction direction)
+xkb_filter_mod_set_func(struct xkb_state *state,
+                        struct xkb_filter *filter,
+                        const struct xkb_key *key,
+                        enum xkb_key_direction direction)
 {
-    if (key != filter->key)
-        return true;
+    if (key != filter->key) {
+        filter->action.mods.flags &= ~ACTION_LOCK_CLEAR;
+        return XKB_FILTER_CONTINUE;
+    }
 
     if (direction == XKB_KEY_DOWN) {
         filter->refcnt++;
-        return false;
+        return XKB_FILTER_CONSUME;
+    }
+    else if (--filter->refcnt > 0) {
+        return XKB_FILTER_CONSUME;
     }
-    if (--filter->refcnt > 0)
-        return false;
 
-    state->clear_mods |= filter->action.mods.mods.mask;
-    if (!(filter->action.mods.flags & ACTION_LOCK_NO_UNLOCK))
-        state->components.locked_mods &= ~filter->priv;
+    state->clear_mods = filter->action.mods.mods.mask;
+    if (filter->action.mods.flags & ACTION_LOCK_CLEAR)
+        state->components.locked_mods &= ~filter->action.mods.mods.mask;
 
     filter->func = NULL;
-    return true;
+    return XKB_FILTER_CONTINUE;
 }
 
 static void
@@ -398,6 +389,30 @@ xkb_filter_mod_lock_new(struct xkb_state *state, struct xkb_filter *filter)
         state->components.locked_mods |= filter->action.mods.mods.mask;
 }
 
+static bool
+xkb_filter_mod_lock_func(struct xkb_state *state,
+                         struct xkb_filter *filter,
+                         const struct xkb_key *key,
+                         enum xkb_key_direction direction)
+{
+    if (key != filter->key)
+        return XKB_FILTER_CONTINUE;
+
+    if (direction == XKB_KEY_DOWN) {
+        filter->refcnt++;
+        return XKB_FILTER_CONSUME;
+    }
+    if (--filter->refcnt > 0)
+        return XKB_FILTER_CONSUME;
+
+    state->clear_mods |= filter->action.mods.mods.mask;
+    if (!(filter->action.mods.flags & ACTION_LOCK_NO_UNLOCK))
+        state->components.locked_mods &= ~filter->priv;
+
+    filter->func = NULL;
+    return XKB_FILTER_CONTINUE;
+}
+
 enum xkb_key_latch_state {
     NO_LATCH,
     LATCH_KEY_DOWN,
@@ -421,6 +436,13 @@ xkb_action_breaks_latch(const union xkb_action *action)
     }
 }
 
+static void
+xkb_filter_mod_latch_new(struct xkb_state *state, struct xkb_filter *filter)
+{
+    filter->priv = LATCH_KEY_DOWN;
+    state->set_mods = filter->action.mods.mods.mask;
+}
+
 static bool
 xkb_filter_mod_latch_func(struct xkb_state *state,
                           struct xkb_filter *filter,
@@ -452,14 +474,14 @@ xkb_filter_mod_latch_func(struct xkb_state *state,
             filter->key = key;
             state->components.latched_mods &= ~filter->action.mods.mods.mask;
             /* XXX beep beep! */
-            return false;
+            return XKB_FILTER_CONSUME;
         }
         else if (xkb_action_breaks_latch(action)) {
             /* XXX: This may be totally broken, we might need to break the
              *      latch in the next run after this press? */
             state->components.latched_mods &= ~filter->action.mods.mods.mask;
             filter->func = NULL;
-            return true;
+            return XKB_FILTER_CONTINUE;
         }
     }
     else if (direction == XKB_KEY_UP && key == filter->key) {
@@ -499,14 +521,7 @@ xkb_filter_mod_latch_func(struct xkb_state *state,
 
     filter->priv = latch;
 
-    return true;
-}
-
-static void
-xkb_filter_mod_latch_new(struct xkb_state *state, struct xkb_filter *filter)
-{
-    filter->priv = LATCH_KEY_DOWN;
-    state->set_mods = filter->action.mods.mods.mask;
+    return XKB_FILTER_CONTINUE;
 }
 
 static const struct {
@@ -538,17 +553,19 @@ xkb_filter_apply_all(struct xkb_state *state,
 {
     struct xkb_filter *filter;
     const union xkb_action *action;
-    bool send = true;
+    bool consumed;
 
     /* First run through all the currently active filters and see if any of
-     * them have claimed this event. */
+     * them have consumed this event. */
+    consumed = false;
     darray_foreach(filter, state->filters) {
         if (!filter->func)
             continue;
-        send = filter->func(state, filter, key, direction) && send;
-    }
 
-    if (!send || direction == XKB_KEY_UP)
+        if (filter->func(state, filter, key, direction) == XKB_FILTER_CONSUME)
+            consumed = true;
+    }
+    if (consumed || direction == XKB_KEY_UP)
         return;
 
     action = xkb_key_get_action(state, key);
@@ -567,9 +584,6 @@ xkb_filter_apply_all(struct xkb_state *state,
         return;
 
     filter = xkb_filter_new(state);
-    if (!filter)
-        return; /* WSGO */
-
     filter->key = key;
     filter->func = filter_action_funcs[action->type].func;
     filter->action = *action;
