@@ -467,7 +467,7 @@ HandleLedNameDef(KeyNamesInfo *info, LedNameDef *def,
 
     if (!ExprResolveString(info->ctx, def->name, &name)) {
         char buf[20];
-        snprintf(buf, sizeof(buf), "%d", def->ndx);
+        snprintf(buf, sizeof(buf), "%u", def->ndx);
         info->errorCount++;
         return ReportBadType(info->ctx, "indicator", "name", buf, "string");
     }
@@ -515,7 +515,7 @@ HandleKeycodesFile(KeyNamesInfo *info, XkbFile *file, enum merge_mode merge)
 
         if (info->errorCount > 10) {
             log_err(info->ctx, "Abandoning keycodes file \"%s\"\n",
-                    file->topName);
+                    file->name);
             break;
         }
     }
