@@ -34,20 +34,21 @@
 #include "test.h"
 
 #include <wayland-client.h>
-#include "xdg-shell-unstable-v5-client-protocol.h"
+#include "xdg-shell-unstable-v6-client-protocol.h"
 #include <wayland-util.h>
 
 struct interactive_dpy {
 	struct wl_display *dpy;
 	struct wl_compositor *compositor;
-	struct xdg_shell *shell;
+	struct zxdg_shell_v6 *shell;
 	struct wl_shm *shm;
 	uint32_t shm_format;
 
 	struct xkb_context *ctx;
 
 	struct wl_surface *wl_surf;
-	struct xdg_surface *xdg_surf;
+	struct zxdg_surface_v6 *xdg_surf;
+	struct zxdg_toplevel_v6 *xdg_top;
 
 	struct wl_list seats;
 };
@@ -70,6 +71,16 @@ struct interactive_seat {
 
 static bool terminate;
 
+#ifdef HAVE_MKOSTEMP
+static int
+create_tmpfile_cloexec(char *tmpname)
+{
+	int fd = mkostemp(tmpname, O_CLOEXEC);
+	if (fd >= 0)
+		unlink(tmpname);
+	return fd;
+}
+#else
 /* The following utility functions are taken from Weston's
  * shared/os-compatibility.c. */
 static int
@@ -103,22 +114,14 @@ set_cloexec_or_close(int fd)
 static int
 create_tmpfile_cloexec(char *tmpname)
 {
-	int fd;
-
-#ifdef HAVE_MKOSTEMP
-	fd = mkostemp(tmpname, O_CLOEXEC);
-	if (fd >= 0)
-		unlink(tmpname);
-#else
-	fd = mkstemp(tmpname);
+	int fd = mkstemp(tmpname);
 	if (fd >= 0) {
 		fd = set_cloexec_or_close(fd);
 		unlink(tmpname);
 	}
-#endif
-
 	return fd;
 }
+#endif
 
 /*
  * Create a new, unique, anonymous file of the given size, and
@@ -141,7 +144,7 @@ create_tmpfile_cloexec(char *tmpname)
  * If posix_fallocate() is not supported, program may receive
  * SIGBUS on accessing mmap()'ed file contents instead.
  */
-int
+static int
 os_create_anonymous_file(off_t size)
 {
 	static const char template[] = "/weston-shared-XXXXXX";
@@ -220,6 +223,9 @@ buffer_create(struct interactive_dpy *inter, uint32_t width, uint32_t height)
 	case WL_SHM_FORMAT_BGR565:
 		stride = width * 2;
 		break;
+        default:
+		fprintf(stderr, "Unsupported SHM format %d\n", inter->shm_format);
+		exit(1);
 	}
 
 	size = stride * height;
@@ -255,56 +261,65 @@ buffer_create(struct interactive_dpy *inter, uint32_t width, uint32_t height)
 }
 
 static void
-surface_configure(void *data, struct xdg_surface *surface,
-                  int32_t width, int32_t height, struct wl_array *states,
+surface_configure(void *data, struct zxdg_surface_v6 *surface,
 		  uint32_t serial)
 {
 	struct interactive_dpy *inter = data;
 
-	if (width == 0 || height == 0) {
-		xdg_surface_ack_configure(inter->xdg_surf, serial);
-		return;
-	}
-
-	buffer_create(inter, width, height);
-	xdg_surface_ack_configure(inter->xdg_surf, serial);
+	zxdg_surface_v6_ack_configure(inter->xdg_surf, serial);
 	wl_surface_commit(inter->wl_surf);
 }
 
+static const struct zxdg_surface_v6_listener surface_listener = {
+	surface_configure,
+};
+
 static void
-surface_close(void *data, struct xdg_surface *surface)
+toplevel_configure(void *data, struct zxdg_toplevel_v6 *toplevel,
+		   int32_t width, int32_t height, struct wl_array *states)
+{
+	struct interactive_dpy *inter = data;
+
+	if (width == 0)
+		width = 200;
+	if (height == 0)
+		height = 200;
+
+	buffer_create(inter, width, height);
+}
+
+static void
+toplevel_close(void *data, struct zxdg_toplevel_v6 *toplevel)
 {
 	terminate = true;
 }
 
-static const struct xdg_surface_listener surface_listener = {
-	surface_configure,
-	surface_close
+static const struct zxdg_toplevel_v6_listener toplevel_listener = {
+	toplevel_configure,
+	toplevel_close
 };
 
 static void surface_create(struct interactive_dpy *inter)
 {
-	int width = 200, height = 200;
-
 	inter->wl_surf = wl_compositor_create_surface(inter->compositor);
-	inter->xdg_surf = xdg_shell_get_xdg_surface(inter->shell,
-	                                            inter->wl_surf);
-	xdg_surface_add_listener(inter->xdg_surf, &surface_listener, inter);
-	xdg_surface_set_title(inter->xdg_surf, "xkbcommon event tester");
-	xdg_surface_set_app_id(inter->xdg_surf,
-	                       "org.xkbcommon.test.interactive-wayland");
-
-	buffer_create(inter, width, height);
+	inter->xdg_surf = zxdg_shell_v6_get_xdg_surface(inter->shell,
+	                                                inter->wl_surf);
+	zxdg_surface_v6_add_listener(inter->xdg_surf, &surface_listener, inter);
+	inter->xdg_top = zxdg_surface_v6_get_toplevel(inter->xdg_surf);
+	zxdg_toplevel_v6_add_listener(inter->xdg_top, &toplevel_listener, inter);
+	zxdg_toplevel_v6_set_title(inter->xdg_top, "xkbcommon event tester");
+	zxdg_toplevel_v6_set_app_id(inter->xdg_top,
+				    "org.xkbcommon.test.interactive-wayland");
 	wl_surface_commit(inter->wl_surf);
 }
 
 static void
-shell_ping(void *data, struct xdg_shell *shell, uint32_t serial)
+shell_ping(void *data, struct zxdg_shell_v6 *shell, uint32_t serial)
 {
-	xdg_shell_pong(shell, serial);
+	zxdg_shell_v6_pong(shell, serial);
 }
 
-static const struct xdg_shell_listener shell_listener = {
+static const struct zxdg_shell_v6_listener shell_listener = {
 	shell_ping
 };
 
@@ -332,7 +347,7 @@ kbd_keymap(void *data, struct wl_keyboard *wl_kbd, uint32_t format,
 	}
 
 	seat->state = xkb_state_new(seat->keymap);
-	if (!seat->keymap) {
+	if (!seat->state) {
 		fprintf(stderr, "Failed to create XKB state!\n");
 		return;
 	}
@@ -418,7 +433,7 @@ pointer_button(void *data, struct wl_pointer *wl_pointer, uint32_t serial,
 {
 	struct interactive_seat *seat = data;
 
-	xdg_surface_move(seat->inter->xdg_surf, seat->wl_seat, serial);
+	zxdg_toplevel_v6_move(seat->inter->xdg_top, seat->wl_seat, serial);
 }
 
 static void
@@ -569,12 +584,11 @@ registry_global(void *data, struct wl_registry *registry, uint32_t name,
 	if (strcmp(interface, "wl_seat") == 0) {
 		seat_create(inter, registry, name, version);
 	}
-	else if (strcmp(interface, "xdg_shell") == 0) {
+	else if (strcmp(interface, "zxdg_shell_v6") == 0) {
 		inter->shell = wl_registry_bind(registry, name,
-		                                &xdg_shell_interface,
+		                                &zxdg_shell_v6_interface,
 		                                MAX(version, 1));
-		xdg_shell_add_listener(inter->shell, &shell_listener, inter);
-		xdg_shell_use_unstable_version(inter->shell, 5);
+		zxdg_shell_v6_add_listener(inter->shell, &shell_listener, inter);
 	}
 	else if (strcmp(interface, "wl_compositor") == 0) {
 		inter->compositor = wl_registry_bind(registry, name,
@@ -615,11 +629,13 @@ dpy_disconnect(struct interactive_dpy *inter)
 		seat_destroy(seat);
 
 	if (inter->xdg_surf)
-		xdg_surface_destroy(inter->xdg_surf);
+		zxdg_surface_v6_destroy(inter->xdg_surf);
+	if (inter->xdg_top)
+		zxdg_toplevel_v6_destroy(inter->xdg_top);
 	if (inter->wl_surf)
 		wl_surface_destroy(inter->wl_surf);
 	if (inter->shell)
-		xdg_shell_destroy(inter->shell);
+		zxdg_shell_v6_destroy(inter->shell);
 	if (inter->compositor)
 		wl_compositor_destroy(inter->compositor);
 	if (inter->shm)
@@ -674,6 +690,7 @@ main(int argc, char *argv[])
 		        (inter.shell) ? "" : "xdg_shell ",
 			(inter.shm) ? "" : "wl_shm",
 			(inter.compositor) ? "" : "wl_compositor");
+                ret = -1;
 		goto err_conn;
 	}
 
