@@ -236,7 +236,8 @@ process_event(xcb_generic_event_t *gevent, struct keyboard *kbd)
         xcb_key_press_event_t *event = (xcb_key_press_event_t *) gevent;
         xkb_keycode_t keycode = event->detail;
 
-        test_print_keycode_state(kbd->state, NULL, keycode);
+        test_print_keycode_state(kbd->state, NULL, keycode,
+                                 XKB_CONSUMED_MODE_XKB);
 
         /* Exit on ESC. */
         if (keycode == 9)
@@ -275,6 +276,10 @@ loop(xcb_connection_t *conn, struct keyboard *kbd)
         }
 
         event = xcb_wait_for_event(conn);
+        if (!event) {
+            continue;
+        }
+
         process_event(event, kbd);
         free(event);
     }
@@ -372,9 +377,9 @@ main(int argc, char *argv[])
         goto err_core_kbd;
     }
 
-    system("stty -echo");
+    test_disable_stdin_echo();
     ret = loop(conn, &core_kbd);
-    system("stty echo");
+    test_enable_stdin_echo();
 
 err_core_kbd:
     deinit_kbd(&core_kbd);
