@@ -22,8 +22,8 @@
  */
 
 #include "test.h"
-#include "xkbcomp-priv.h"
-#include "rules.h"
+#include "xkbcomp/xkbcomp-priv.h"
+#include "xkbcomp/rules.h"
 
 struct test_data {
     /* Rules file */
